# Copyright (c) 2019, TU Wien, Department of Geodesy and Geoinformation (GEO).
# All rights reserved.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from libc.math cimport exp
import numpy as np
cimport numpy as np
cimport cython


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def swi_calc_cy(np.ndarray[np.double_t] juldate,
                np.ndarray[np.float32_t] ssm,
                np.ndarray[np.int_t] ctime,
                np.ndarray[np.double_t] swi_jd,
                np.ndarray[np.float_t] nom,
                np.ndarray[np.float_t] denom,
                double last_jd_var,
                np.ndarray[np.float_t] norm_factor,
                double nan):

    cdef int i
    cdef int j = 0
    cdef int k
    cdef np.ndarray[np.float_t] ef = np.empty(len(ctime))
    cdef double tdiff
    cdef float qflag
    cdef int len_swi = len(swi_jd)
    cdef int len_jd = len(juldate)
    cdef int len_ctime = len(ctime)
    cdef np.ndarray[np.float32_t, ndim = 2] swi_ts = np.empty((len_swi, len_ctime), dtype=np.float32)
    cdef np.ndarray[np.float32_t, ndim = 2] swi_qflag = np.empty((len_swi, len_ctime), dtype=np.float32)

    for i in range(len_swi):

        while j < len_jd and juldate[j] <= swi_jd[i]:
            if ssm[j] != nan or ssm[j] == np.nan:
                tdiff = juldate[j] - last_jd_var
                for k in range(len_ctime):
                    ef[k] = exp(-tdiff / ctime[k])
                    nom[k] = ef[k] * nom[k] + ssm[j]
                    denom[k] = ef[k] * denom[k] + 1

                last_jd_var = juldate[j]
            j += 1

        if denom[0] == 0:
            continue  # no valid SSM measurement before swi_jd[i]

        for k in range(len_ctime):
            if ssm[i] != nan or ssm[i] == np.nan:
                swi_ts[i, k] = nom[k] / denom[k]
            else:
                swi_ts[i, k] = np.nan

            qflag = 100.0 * \
                denom[k] * exp(-(swi_jd[i] - last_jd_var) /
                               ctime[k]) / norm_factor[k]
            if qflag > 100:
                qflag = 100
            swi_qflag[i, k] = qflag

    return swi_ts, swi_qflag, nom, denom, last_jd_var


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def swi_calc_cy_noise(np.ndarray[np.double_t] juldate,
                      np.ndarray[np.float32_t] ssm,
                      np.ndarray[np.int_t] ctime,
                      np.ndarray[np.double_t] swi_jd,
                      np.ndarray[np.float_t] nom,
                      np.ndarray[np.float_t] denom,
                      double last_jd_var,
                      np.ndarray[np.float32_t] ssm_noise,
                      np.ndarray[np.double_t] nom_noise):

    cdef int i = 0
    cdef int j = 0
    cdef int k = 0
    cdef np.ndarray[np.float_t] ef = np.empty(len(ctime))
    cdef double tdiff = 0.0
    cdef int len_swi = len(swi_jd)
    cdef int len_jd = len(juldate)
    cdef int len_ctime = len(ctime)
    cdef np.ndarray[np.float32_t, ndim = 2] swi_ts = np.empty((len_swi, len_ctime), dtype=np.float32)
    cdef np.ndarray[np.float32_t, ndim = 2] swi_noise = np.empty((len_swi, len_ctime), dtype=np.float32)

    for i in range(len_swi):

        while j < len_jd and juldate[j] <= swi_jd[i]:
            tdiff = juldate[j] - last_jd_var
            for k in range(len_ctime):
                ef[k] = exp(-tdiff / ctime[k])
                nom[k] = ef[k] * nom[k] + ssm[j]
                denom[k] = ef[k] * denom[k] + 1
                exp_term = ef[k] ** 2
                nom_noise[k] = nom_noise[k] * exp_term + ssm_noise[j]
            last_jd_var = juldate[j]
            j += 1

        if denom[0] == 0:
            continue  # no valid SSM measurement before swi_jd[i]

        for k in range(len_ctime):
            swi_ts[i, k] = nom[k] / denom[k]
            denom_ns = denom[k] ** 2
            swi_noise[i, k] = nom_noise[k] / denom_ns

    return swi_ts, swi_noise, nom, denom, last_jd_var, nom_noise



# Copyright (c) 2017, Vienna University of Technology (TU Wien),
# Department of Geodesy and Geoinformation (GEO).
# All rights reserved.
#
# All information contained herein is, and remains the property of Vienna
# University of Technology (TU Wien), Department of Geodesy and Geoinformation
# (GEO). The intellectual and technical concepts contained herein are
# proprietary to Vienna University of Technology (TU Wien), Department of
# Geodesy and Geoinformation (GEO). Dissemination of this information or
# reproduction of this material is forbidden unless prior written permission
# is obtained from Vienna University of Technology (TU Wien), Department of
# Geodesy and Geoinformation (GEO).

'''
Test for the iteration data storage
'''

import os
import unittest
import glob
import tempfile
import numpy.testing as nptest
import numpy as np
from datetime import datetime

from pyswi.swi_img.iterative_storage import IterStepData


def curpath():
    pth, _ = os.path.split(os.path.abspath(__file__))
    return pth


class TestIterStepData(unittest.TestCase):

    def setUp(self):

        self.itersavepath = tempfile.mkdtemp()

    def test_metadata_extraction(self):
        path = os.path.join(curpath(), 'data', 'iterative_storage', 'metadata_from_filename')
        data_coll = IterStepData(path, 100, {})

        assert data_coll.metadata['sensing_start'] == [
            datetime(2013, 11, 24, 12)]
        assert data_coll.metadata['sensing_end'] == [
            datetime(2013, 11, 24, 12, 2, 58)]
        assert data_coll.metadata['processing_end'] == [
            datetime(2013, 11, 25, 17, 47, 33)]

    def test_get_latest_sensing_time(self):
        path = os.path.join(curpath(), 'data', 'iterative_storage', 'metadata_from_filename')
        data_coll = IterStepData(path, 100, {})

        max_sensing_time = data_coll.get_latest_sensing_time()
        assert max_sensing_time == datetime(2013, 11, 24, 12, 2, 58)

    def test_construct_filename(self):
        path = os.path.join(curpath(), 'data', 'iterative_storage', 'metadata_from_filename')
        data_coll = IterStepData(path, 100, {}, prefix='ASCA_SWI25_03_M02')

        filename = data_coll._construct_file_name(datetime(2013, 11, 24, 12),
                                                  datetime(
                                                      2013, 11, 24, 12, 2, 58),
                                                  datetime(2013, 11, 25, 17, 47, 33))

        assert filename == 'ASCA_SWI25_03_M02_20131124120000Z_20131124120258Z_20131125174733Z.nc'

    def test_iter_data_read_write(self):

        data_coll = IterStepData(self.itersavepath, 100,
                                 variables={'gain_sigma': -9999.,
                                            'ssf': 255})
        iter_data = data_coll.get_empty_data()
        assert iter_data['ssf'][0] == 255
        assert iter_data['ssf'].dtype == int
        assert iter_data['gain_sigma'][0] == -9999.
        assert iter_data['gain_sigma'].dtype == np.float64
        iter_data['ssf'][50] = 38
        iter_data['gain_sigma'][50] = 48
        iter_data['header']['sensing_start'] = datetime(2007, 1, 1, 12, 15, 22)
        iter_data['header']['sensing_end'] = datetime(2007, 1, 2, 11, 59, 58)
        iter_data['header']['processing_end'] = datetime(2014, 1, 1, 15, 20, 43)
        iter_data['header']['processing_start'] = datetime(2014, 1, 1, 14, 20, 43)
        data_coll.save_iter_data(iter_data)
        restored_data = data_coll.read_latest_iter_data()
        restored_header = restored_data.pop('header')
        iter_header = iter_data.pop('header')
        for name in iter_header:
            assert restored_header[name] == iter_header[name]

        for name in restored_header:
            assert restored_header[name] == iter_header[name]

        for name in iter_data:
            nptest.assert_allclose(restored_data[name], iter_data[name])



# Copyright (c) 2021, Vienna University of Technology (TU Wien),
# Department of Geodesy and Geoinformation (GEO).
# All rights reserved.
#
# All information contained herein is, and remains the property of Vienna
# University of Technology (TU Wien), Department of Geodesy and Geoinformation
# (GEO). The intellectual and technical concepts contained herein are
# proprietary to Vienna University of Technology (TU Wien), Department of
# Geodesy and Geoinformation (GEO). Dissemination of this information or
# reproduction of this material is forbidden unless prior written permission
# is obtained from Vienna University of Technology (TU Wien), Department of
# Geodesy and Geoinformation (GEO).

'''
Storage of Iteration data between processing runs.
'''
import numpy as np
import os
import glob
from datetime import datetime
import netCDF4
from trollsift import Parser


class IterStepData(object):

    """
    reads and saves data needed between iteration steps

    The two methods get_empty_data and save_iter_data can be overwritten
    if more complex data structures are necessary.
    By default one dimensional arrays of size n_obs and a simple header are used.

    The default header is:

        header = {'product_name': "",
                  'parent_product_name': "",
                  'instrument_id': "",
                  'sensing_start': datetime(1900, 1, 1),
                  'sensing_end': datetime(4100, 12, 31, 15, 0, 0),
                  'processing_end': ""}

    Parameters
    ----------
    path : string
        path to which the iteration step data should be read from
        and saved to
    n_obs: int
        number of observations to store for the full target grid
    variables: dict
        a mapping from variable name to nan/fill value to use.
        This sets the datatype and initial values for the variable
        that is stored.
    prefix: string, optional
        filename prefix
    """

    def __init__(self, path, n_obs,
                 variables,
                 prefix='GENERIC_PREFIX'):
        self.path = path
        self.n_obs = n_obs
        self.variables = variables
        self.prefix = prefix
        self.files = 0
        self.metadata = None
        self.datestoragefmt = "%Y-%m-%d %H:%M:%S"
        self.fname_datefmt = '%Y%m%d%H%M%SZ'
        self.fname_template = '_'.join(
            [self.prefix,
             "{{sensing_start:{}}}".format(self.fname_datefmt),
             "{{sensing_end:{}}}".format(self.fname_datefmt),
             "{{processing_end:{}}}.nc".format(self.fname_datefmt)])
        self.fname_parser = Parser(self.fname_template)
        self._load_info()

    def _construct_file_name(self, sensing_start, sensing_end,
                             processing_end):
        """
        construct filename from sensing start and end and
        processing end

        Parameters
        ----------
        sensing_start : datetime
            time of observation of first measurement in file
        sensing_end : datetime
            time of observation of last measurement in file
        processing_end : datetime
            processing timestamp

        Returns
        -------
        filename : string
            filename
        """

        return self.fname_parser.compose({'sensing_start': sensing_start,
                                          'sensing_end': sensing_end,
                                          'processing_end': processing_end})

    def _load_info(self):
        """
        load the information about the files from the path
        """
        self.files = self.find_files()
        if len(self.files) == 0:
            self.files_available = False
        else:
            self.files_available = True
        self.metadata = self.get_metadata()

    def find_files(self):
        """
        find files in path by searching for 'self.prefix*'

        Returns
        -------
        files : list
            list of full filepaths
        """
        files = glob.glob(os.path.join(self.path, ''.join([self.prefix, '*'])))
        files.sort()

        return files

    def get_metadata(self):
        """
        loads metadata from filenames. Important field is the end of sensing
        in each file which indicates if a bufr file has already been processed
        """
        metadata = {'sensing_start': [],
                    'sensing_end': [],
                    'processing_end': []}

        fmt = '%Y%m%d%H%M%S'

        for _file in self.files:
            filename = os.path.split(_file)[1]
            mdata = self.fname_parser.parse(filename)
            metadata['sensing_start'].append(mdata['sensing_start'])
            metadata['sensing_end'].append(mdata['sensing_end'])
            metadata['processing_end'].append(mdata['processing_end'])

        return metadata

    def _get_latest_idx(self):
        """
        get index of latest sensing file
        """
        if type(self.metadata) != dict:
            self.get_metadata()

        max_sensing_idx = np.argmax(self.metadata['sensing_end'])

        return max_sensing_idx

    def get_latest_sensing_time(self):
        """
        find the latest time of the observations the iteration files in the
        folder already contain
        """

        if type(self.metadata) != dict:
            self.get_metadata()

        max_sensing_time = self.metadata['sensing_end'][self._get_latest_idx()]

        return max_sensing_time

    def read_latest_iter_data(self):
        """
        reads the latest iteration data
        """

        filename = self.files[self._get_latest_idx()]
        structure = {'header': {}}
        with netCDF4.Dataset(filename, 'r') as nc:
            for variable in nc.variables:
                try:
                    structure[variable] = nc.variables[variable][:].filled()
                except AttributeError:
                    structure[variable] = nc.variables[variable][:]
            for name in nc.ncattrs():
                structure['header'][name] = getattr(nc, name)

            from_string = ['sensing_start', 'sensing_end',
                           'processing_start', 'processing_end']
            for name in from_string:
                structure['header'][name] = datetime.strptime(
                    structure['header'][name], self.datestoragefmt)

        return structure

    def get_empty_data(self):
        """
        returns a empty structure dictionary
        """

        header = {'product_name': "",
                  'parent_product_name': "",
                  'instrument_id': "",
                  'sensing_start': datetime(1900, 1, 1),
                  'sensing_end': datetime(4100, 12, 31, 15, 0, 0),
                  'processing_end': ""}

        structure = {'header': header}
        for variable in self.variables:
            nan_array = np.empty(
                self.n_obs, dtype=type(self.variables[variable]))
            nan_array.fill(self.variables[variable])
            structure[variable] = nan_array

        return structure

    def save_iter_data(self, data, refresh_coll=True):
        """
        save iteration data to file and refresh the collection if set

        Parameters
        ----------
        data : dict
            dictionary in the format given by get_empty_data
        refresh_coll : boolean
            if True then the collection is refreshed after the file is saved
            which is good if processing should continue with new iteration
            data
        """
        filename = self._construct_file_name(data['header']['sensing_start'],
                                             data['header']['sensing_end'],
                                             data['header']['processing_end'])

        with netCDF4.Dataset(os.path.join(self.path, filename), 'w') as nc:
            nc.createDimension('observations', self.n_obs)
            header = data['header'].copy()
            # convert date object to string for storage in netCDF
            to_string = ['sensing_start', 'sensing_end',
                         'processing_start', 'processing_end']
            for name in to_string:
                header[name] = header[name].strftime(self.datestoragefmt)

            nc.setncatts(header)
            for variable in self.variables:
                var = nc.createVariable(variable, data[variable].dtype, ('observations'),
                                        fill_value=self.variables[variable], zlib=True)
                var[:] = data[variable]

        if refresh_coll:
            self._load_info()



# Copyright (c) 2021, TU Wien, Department of Geodesy and Geoinformation (GEO).
# All rights reserved.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import os
from datetime import datetime

import numpy as np
from pyswi.swi_img.iterative_storage import IterStepData
from cadati.jd_date import julian2datetime

def calc_swi(next_ssm, next_ssm_jd, tvalue, swi, jd,
                  gain=None, weighted=False, den=None, n=None, wsum=None, next_w=None):

    """
    Takes input data and previous data and calculates swi_img
    values for the next date. All arrays have to be the same shape
    so that the whole batch can be calculated at once
    initializing and masking has to be done outside of this basic
    function

    Parameters
    ----------
    next_ssm : numpy.array
        surface soil moisture
    next_ssm_jd : numpy.array
        observation time of ssm given as julian date
    tvalue : int
        T-value for SWI calculation
    swi : numpy.array
        values of SWI at previous iteration
    jd : numpy.array
        values of julian date from previous iteration
    gain : numpy.array (optional, default=None)
        values of gain from previous iteration. Required for non-weighted swi calculation.
    weighted : boolean (default=False)
        Weighted swi calculation following equation (9) in the paper:
        Bauer-Marschallinger B, Paulik C, Hochstoeger S, Mistelbauer T,
        Modanesi S, Ciabatta L, Massari C, Brocca L, Wagner W.
        "Soil moisture from fusion of scatterometer and SAR:
        Closing the scale gap with temporal filtering."
        Remote Sensing. 2018 Jul;10(7):1030.
    den: numpy.array (optional, default=None)
        values of denominator from previous iteration. Required for weighted swi calculation.
    n : numpy.array (optional, default=None)
        values of total number of previous iterations. Required for weighted swi calculation.
    wsum : numpy.array (optional, default=None)
        values of sum of weights applied in previous iteration. Required for weighted swi calculation.
    next_w : numpy.array (optional, default=None)
        weights given to ssm values. Required for weighted swi calculation.

    "next" refers to t_(i+1)
    "previous" refers to t_(i)

    Returns
    -------
    next_swi : numpy.array
        swi results
    next_den : numpy.array
        denominator for next iteration
    next_n : numpy.array
        total number of iterations next iteration
    next_wsum : numpy.array
        sum of weights for next iteration
    next_gain : numpy.array
        gain for next iteration
    """

    # Calculate time diff
    time_diff = (next_ssm_jd - jd).astype(np.float64)

    if weighted:

        next_e_term = np.exp(-(time_diff / float(tvalue)))

        # calculate denominator based on previous values
        next_den = (1.0 + next_e_term * den).astype(np.float32)

        # increase total number of iterations (i=0, regularly)
        next_n = (n + 1.0).astype(np.int32)

        # calculate next sum of weights
        next_wsum = (next_w + wsum).astype(np.float32)

        # calculate weighted SWI
        next_swi = ((swi * next_n / n * wsum * (next_den - 1) +
                     next_ssm * next_n * next_w) /
                    (next_wsum * next_den)).astype(np.float32)

        # rectify overshooting due to weighting
        next_swi[np.where(next_swi < 0)] = 0
        next_swi[np.where(next_swi > 100)] = 100

        return next_swi, next_den, next_n, next_wsum

    else:

        # calculate gain
        next_gain = (gain / (gain.astype(np.float64) +
                             np.exp(-(time_diff / float(tvalue)))))
        # calculate SWI
        next_swi = swi + next_gain * (next_ssm - swi)

        return next_swi, next_gain


class IterativeMultiSWI(object):
    """Class for calculation of the SWI in iterative mode.

    This means that previous_swi, previous_gain, previous_qflag and
    previous_julian_date have to be stored for a successful
    restart after a break.

    This class depends on the IterativeSWI class, but takes multiple T-values
    as argument.

    This is suitable if the SWI is calculated from a series of swaths/images.

    Parameters
    ----------
    ssm: numpy.ndarray
        An array of the soil moisture values.
    iter_data_path: string
        Path where the iteration data is/will be stored.
    tvalues: list
        List of tvalue integer values used for SWI calculation.
    weighted: boolean (default=False)
        Weighted swi calculation following equation (9) in the paper:
        Bauer-Marschallinger B, Paulik C, Hochstoeger S, Mistelbauer T,
        Modanesi S, Ciabatta L, Massari C, Brocca L, Wagner W.
        "Soil moisture from fusion of scatterometer and SAR:
        Closing the scale gap with temporal filtering."
        Remote Sensing. 2018 Jul;10(7):1030
    """

    def __init__(self, ssm, iter_data_path, tvalues, weighted=False):
        """
        Initializes the IterativeSWI class.

        Parameters
        ----------
        ssm: numpy.ndarray
            Soil Moisture values of the whole image.
        iter_data_path: string
            Output and input path of the calculation results.
        tvalues: list
            List of the tvalue integer values used for SWI calculation.
        weighted: boolean (default=False)
            Weighted swi calculation following equation (9) in the paper:
            Bauer-Marschallinger B, Paulik C, Hochstoeger S, Mistelbauer T,
            Modanesi S, Ciabatta L, Massari C, Brocca L, Wagner W.
            "Soil moisture from fusion of scatterometer and SAR:
            Closing the scale gap with temporal filtering."
            Remote Sensing. 2018 Jul;10(7):1030.
        """
        iter_swi = []
        prev_iter_files = []

        for i in range(0, len(tvalues)):
            it_swi = IterativeSWI(ssm, iter_data_path, tvalues[i], weighted=weighted)
            iter_swi.append(it_swi)
            prev_iter_files.append(it_swi.iterstepdata.files)

        self.calc_swi = iter_swi
        self.prev_iter_files = prev_iter_files
        self.tvalues = tvalues

    def load_iter_data(self):
        """
        Loads the iterative swi_img data for all tvalue values
        from the iter_data_path to iter_data.
        If there are no feasible files, the iter_data is set empty.
        """
        for it_data in self.calc_swi:
            it_data.load_iter_data()

    def store_iter_data(self):
        """ Stores the iterative swi_img data to the iter_data_path. """
        for it_data in self.calc_swi:
            it_data.store_iter_data()

    def remove_prev_iter_files(self):
        """ Removes the previous iter files in the iter_data_path. """
        for it_file_list in self.prev_iter_files:
            for it_file in it_file_list:
                os.remove(it_file)

    def calc_iter(self, next_ssm_jd, next_ssm, ind=None, weighted=False, next_w=None):
        """
        Calculate SWI iteratively.

        Parameters
        ----------
        next_ssm_jd: numpy.ndarray
            observation time of ssm given as julian date
        next_ssm: numpy.ndarray
            Soil moisture values
        ind: numpy.ndarray (optional, default=None)
            Index of valid next_ssm pixels.
            By default the complete image array is used.
        weighted : boolean (default=False)
            Weighted swi calculation following Equation (9) in the paper:
            Bauer-Marschallinger B, Paulik C, Hochstoeger S, Mistelbauer T,
            Modanesi S, Ciabatta L, Massari C, Brocca L, Wagner W.
            "Soil moisture from fusion of scatterometer and SAR:
            Closing the scale gap with temporal filtering."
            Remote Sensing. 2018 Jul;10(7):1030
        next_w : numpy.ndarray (optional, default=None)
            Weights given to ssm values.

        Returns
        ------
        swi_dict: dict
            dict with the new SWI values for all tvalue values.
            The keys of the dictionary are saved like e.g. SWI_001.
        """

        swi_dict = {}

        for i in range(0, len(self.tvalues)):
            key_swi = "SWI_%03d" % (self.tvalues[i],)
            swi_dict[key_swi] = self.calc_swi[i].calc_iter(next_ssm_jd, next_ssm, ind=ind,
                                                                weighted=weighted, next_w=next_w)

        return swi_dict


class IterativeSWI(object):
    """
    Class for calculation of the SWI in iterative mode.

    This means that the previous: swi, julian dates, qflags and/or gain, denominator, weights
    have to be stored for a successful restart after a break. On the first iteration the swi
    values are initialized as equal to the surface soil moisture, and gain is initialized as 1.0
    following Albergel et al (2008) https://doi.org/10.5194/hess-12-1323-2008

    This class only works with a single tvalue value.

    This is suitable if the SWI is calculated from a series of swaths/images.

    Parameters
    ----------
    ssm: numpy.ndarray
        Soil Moisture values of the whole image.
    iter_data_path: string
        Path to where the iteration data is/will be stored.
    tvalue: int
        T-value for SWI calculation
    weighted : boolean (default=False)
        Weighted swi calculation following equation (9) in the paper:
        Bauer-Marschallinger B, Paulik C, Hochstoeger S, Mistelbauer T,
        Modanesi S, Ciabatta L, Massari C, Brocca L, Wagner W.
        "Soil moisture from fusion of scatterometer and SAR:
        Closing the scale gap with temporal filtering."
        Remote Sensing. 2018 Jul;10(7):1030.
    """

    def __init__(self, ssm, iter_data_path, tvalue, weighted=False):
        """
        Initializes the IterativeSWI class.

        Parameters
        ----------
        ssm: numpy.ndarray
            Soil Moisture values of the whole image.
        iter_data_path: string
            Output and input path of the calculation results.
        tvalue: int
            T-value for SWI calculation
        weighted : boolean (default=False)
            Weighted swi calculation following equation (9) in the paper:
            Bauer-Marschallinger B, Paulik C, Hochstoeger S, Mistelbauer T,
            Modanesi S, Ciabatta L, Massari C, Brocca L, Wagner W.
            "Soil moisture from fusion of scatterometer and SAR:
            Closing the scale gap with temporal filtering."
            Remote Sensing. 2018 Jul;10(7):1030.
        """

        pref = "SWI_%03d" % (tvalue,)
        self.iter_data_path = iter_data_path
        self.float_nan = np.float32(np.nan)
        self.jd_nan = np.float64(np.nan)
        if weighted:
            self.iterstepdata = IterStepData(self.iter_data_path,
                                             len(ssm),
                                             {'swi': self.float_nan,
                                              'qflag': 1.0,
                                              'den': self.float_nan,
                                              'n': 0,
                                              'wsum': self.float_nan,
                                              'jd': self.jd_nan},
                                             prefix=pref)
        else:
            self.iterstepdata = IterStepData(self.iter_data_path,
                                             len(ssm),
                                             {'swi': self.float_nan,
                                              'qflag': 1.0,
                                              'gain': self.float_nan,
                                              'jd': self.jd_nan},
                                             prefix=pref)
        self.tvalue = tvalue
        self.processing_start = datetime.now()
        self.load_iter_data()

    def load_iter_data(self):
        """
        Loads the iterative swi_img data from the iter_data_path to iter_data.
        If there is no feasible file, the iter_data is set empty.
        """
        if self.iterstepdata.files_available is False:
            empty_data = self.iterstepdata.get_empty_data()
            empty_data['header']['processing_start'] = datetime(1900, 1, 1)
            empty_data['header']['processing_end'] = datetime(1901, 1, 1)
            self.iter_data = empty_data
        else:
            self.iter_data = self.iterstepdata.read_latest_iter_data()

    def store_iter_data(self):
        """
        Stores the iterative swi_img data to the iter_data_path.
        """
        self.iterstepdata.save_iter_data(self.iter_data)


    def calc_iter(self, next_ssm_jd, next_ssm, ind=None, weighted=False, next_w=None):
        """
        Calculate SWI iteratively.

        Parameters
        ----------
        next_ssm_jd: numpy.ndarray
            Observation time of ssm given as julian date
        next_ssm: numpy.ndarray
            Surface soil moisture values
        ind: numpy.ndarray
            Index of valid next_ssm pixels
        weighted: boolean (default=False)
            Weighted swi calculation following equation (9) in the paper:
            Bauer-Marschallinger B, Paulik C, Hochstoeger S, Mistelbauer T,
            Modanesi S, Ciabatta L, Massari C, Brocca L, Wagner W.
            "Soil moisture from fusion of scatterometer and SAR:
            Closing the scale gap with temporal filtering."
            Remote Sensing. 2018 Jul;10(7):1030.
        next_w : numpy.ndarray (optional, default=None)
            Weights given to ssm values

        Returns
        ------
        swi_img: numpy.ndarray
            Array with the new calculated swi_img values.
        """

        if ind is None:
            ind = np.arange(0, self.iter_data['swi'].size, 1)

        # load values from previous caluclation
        swi = self.iter_data['swi'][ind]
        jd = self.iter_data['jd'][ind]
        if weighted:
            den = self.iter_data['den'][ind]
            n = self.iter_data['n'][ind]
            wsum = self.iter_data['wsum'][ind]
        else:
            gain = self.iter_data['gain'][ind]

        still_untouched = np.isnan(swi)
        swi[still_untouched] = next_ssm[still_untouched]
        jd[still_untouched] = next_ssm_jd[still_untouched]
        if weighted:
            wsum[still_untouched] = next_w[still_untouched]
            den[still_untouched] = 1.0
            n[still_untouched] = 1
        else:
            gain[still_untouched] = 1.0

        # calculate new values
        if weighted:
            next_swi, next_den, next_n, next_wsum = calc_swi(next_ssm, next_ssm_jd,
                                       self.tvalue, swi, jd, den, n, wsum, next_w)
        else:
            next_swi, next_gain = calc_swi(next_ssm, next_ssm_jd, self.tvalue, swi, jd, gain=gain)

        # update values of iter data with now calculated values
        self.iter_data['swi'][ind] = next_swi
        self.iter_data['jd'][ind] = next_ssm_jd
        if weighted:
            self.iter_data['den'][ind] = next_den
            self.iter_data['n'][ind] = next_n
            self.iter_data['wsum'][ind] = next_wsum
        else:
            self.iter_data['gain'][ind] = next_gain

        # update iter_data header for complete file.
        valid_jd = np.isfinite(self.iter_data['jd'])
        header = {'sensing_start': julian2datetime(float(np.nanmin(self.iter_data['jd'][valid_jd]))),
                  'sensing_end': julian2datetime(float(np.max(self.iter_data['jd'][valid_jd]))),
                  'processing_start': self.processing_start,
                  'processing_end': datetime.now()}
        self.iter_data['header'].update(header)

        return self.iter_data['swi']



# Copyright (c) 2021, TU Wien, Department of Geodesy and Geoinformation (GEO).
# All rights reserved.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from math import exp, sqrt

import pyximport
import numpy as np

pyximport.install(setup_args={'include_dirs': [
                  np.get_include()]}, inplace=True)

from pyswi.swi_ts.swi_calc_routines import swi_calc_cy, swi_calc_cy_noise

def swi_error_prop(ssm, t_value, t_noise, swi_error, gain_in=None, nan=-9999.):
    """
    Recursive SWI calculation and error propagation function
    based on DeSantis and Biondi (2018; https://doi.org/10.29007/kvhb)
    Translated from MatLab code obtained from the authors.

    Parameters
    ----------
    ssm: numpy.ndarray
        Surface soil moisture time series with fields 'sm', 'sm_uncertainty' and 'sm_jd'
    t_value: numpy.ndarray
        Exponential filter characteristic T-value parameter
    t_noise: numpy.ndarray
        T-value standard error.
        10% of T for calibrated T-values.
    swi_error: numpy.ndarray
        Exponential filter model structural error.
        ubMSE(ISMNswi, ISMNrzsm), based on empirical experiments.
    gain_in: dict
        stored parameters of the last iteration.
    nan: float
        nan value of the input ssm dataset

    Returns
    -------
    swi: numpy.ndarray
        Soil water index time series
    swi_noise:numpy.ndarray
        Soil water index noise time series
    """

    len_ssm = len(ssm)
    len_T = len(t_value)
    swi = np.zeros((len_ssm, len_T))
    swi_noise = np.zeros((len_ssm, len_T))
    qflag = np.ones((len_ssm, len_T))

    if gain_in is None:
        swi[0] = ssm['sm'][0]
        swi_noise[0] = ssm['sm_uncertainty'][0]
        qflag[0] = 1.
        last_jd = ssm['sm_jd'][0]
        gain_curr = [1] * len_T
        contr1_curr = [ssm['sm_uncertainty'][0]**2] * len_T
        G_curr = [0] * len_T
        JT_curr = [0] * len_T # Jacobian term
    else:
        last_jd = gain_in['last_jd']
        time_diff = ssm['sm_jd'][0] - gain_in['last_jd']
        ef = [np.exp(-time_diff / t_value[i]) for i in range(0, len(t_value))]
        swi[0] = gain_in['last_swi'] + gain_in['last_gain'] * (ssm['sm'][0] - gain_in['last_swi'])
        gain_curr = [gain_in['last_gain'][i] / (gain_in['last_gain'][i] + ef[i]) for i in range(0, len(t_value))]
        contr1_curr = [((1 - gain_curr[i])**2) * gain_in['last_contr1'][i] + (gain_curr[i] * ssm['sm_uncertainty'][0])**2
                       for i in range(0, len(t_value))]
        G_curr = [ef[i] * (gain_in['last_G'][i] + time_diff / t_value[i] / gain_in['last_gain'][i]) for i in range(0, len(t_value))]
        JT_curr = [gain_curr[i] / t_value[i] * (G_curr[i] * (gain_in['last_swi'][i] - swi[0][i]) + ef[i] * t_value[i]
                                                / gain_in['last_gain'][i] * gain_in['last_JT'][i]) for i in range(0, len(t_value))]
        contr2 = [(JT_curr[i] * t_noise[i])**2 for i in range(0, len(t_value))]
        swi_noise[0] = [sqrt(contr1_curr[i] + contr2[i] + swi_error[i]) for i in range(0, len(t_value))]
        qflag[0] = gain_in['last_qflag']
        last_jd += 1

    gain_old = [None] * len_T
    contr1_old = [None] * len_T
    contr2 = [None] * len_T
    G_old = [None] * len_T
    JT_old = [None] * len_T

    j = 0

    for i in range(len_ssm):
        while j < len_ssm and ssm['sm_jd'][j] <= ssm['sm_jd'][i]:
            if ssm['sm'][j] != nan or ssm['sm'][j] == np.nan:
                time_diff = ssm['sm_jd'][j] - last_jd

                for c in range(len_T):
                    ef = np.exp(-time_diff / t_value[c])
                    gain_old[c] = gain_curr[c]
                    contr1_old[c] = contr1_curr[c]
                    gain_curr[c] = gain_old[c] / (gain_old[c] + ef)
                    swi[i][c] = swi[i-int(time_diff)][c] + gain_curr[c] * (ssm['sm'][i] - swi[i-int(time_diff)][c])
                    contr1_curr[c] = ((1 - gain_curr[c])**2) * contr1_old[c] + (gain_curr[c] * ssm['sm_uncertainty'][i])**2
                    G_old[c] = G_curr[c]
                    JT_old[c] = JT_curr[c]
                    G_curr[c] = ef * (G_old[c] + time_diff / t_value[c] / gain_old[c])
                    JT_curr[c] = gain_curr[c] / t_value[c] * (G_curr[c] * (swi[i-int(time_diff)][c] - swi[i][c])
                                                        + ef * t_value[c] / gain_old[c] * JT_old[c])
                    contr2[c] = (JT_curr[c] * t_noise[c])**2
                    swi_noise[i][c] = sqrt(contr1_curr[c] + contr2[c] + swi_error[c])

                last_jd = ssm['sm_jd'][i]
            j += 1

        for c in range(len_T):
            if ssm['sm'][i] != nan or ssm['sm'][i] == np.nan:
                swi[i, c] = swi[i-int(time_diff)][c] + gain_curr[c] * (ssm['sm'][i] - swi[i-int(time_diff)][c])
                swi_noise[i, c] = sqrt(contr1_curr[c] + contr2[c] + swi_error[c])
                qflag[i, c] = 1 + qflag[i-1][c] * np.exp(-(1. / float(t_value[c])))

            else:
                swi[i, c] = np.nan
                swi_noise[i, c] = np.nan
                qflag[i, c] = qflag[i-1][c] * np.exp(-(1. / float(t_value[c])))

    gain_out = {'last_jd': last_jd, 'last_gain': gain_curr, 'last_contr1': contr1_curr,
                'last_G': G_curr, 'last_JT': JT_curr, 'last_swi': swi[-1], 'last_noise': swi_noise[-1], 'last_qflag': qflag[-1]}

    dtype_list = [('swi_jd', np.float64)]
    for t in t_value:
        dtype_list.append(('swi_noise_{}'.format(t), np.float32))
        dtype_list.append(('swi_{}'.format(t), np.float32))
        dtype_list.append(('qflag_{}'.format(t), np.float32))

    swi_ts = np.zeros(ssm.size, dtype=np.dtype(dtype_list))

    swi_ts['swi_jd'] = ssm['sm_jd']
    for i, t in enumerate(t_value):
        swi_ts['swi_{}'.format(t)] = swi[:, i]
        swi_ts['swi_noise_{}'.format(t)] = swi_noise[:, i]
        swi_ts['qflag_{}'.format(t)] = qflag[:, i]

    return swi_ts, gain_out


def calc_swi_ts(ssm_ts, swi_jd, gain_in=None, t_value=[1, 5, 10, 15, 20],
                nom_init=0, denom_init=0, nan=-9999.):
    """
    Time series calculation of the Soil Water Index.

    Parameters
    ----------
    ssm_ts : numpy.ndarray or dict
        Surface soil moisture time series with fields: sm_jd, sm, sm_noise
    swi_jd : numpy.ndarray
        Julian date time stamps of the SWI time series.
    gain_in : dict, optional
        Gain parameters of last calculation.
        Dictionary with fields: last_jd, nom, denom
    t_value : list, optional
        Characteristic time length (default: 1, 5, 10, 15, 20).
    nom_init : float64, optional
        Initial value of nom in the SWI calculation (default: 0).
    denom_init : float64, optional
        Initial value of denom in the SWI calculation (default: 0).
    nan : float64, optional
        NaN value to be masked in the SWI retrieval.

    Returns
    -------
    swi_ts : numpy.ndarray
        Soil Water Index (SWI) time series.
    gain_out : dict
        Gain parameters of last calculation.
        fields gpi, last_jd, nom, denom, nom_ns
    """

    t_value = np.asarray(t_value)

    if gain_in is None:
        gain_in = {'last_jd': np.float64(0),
                   'denom': np.full(len(t_value), denom_init, dtype=np.float64),
                   'nom': np.full(len(t_value), nom_init, dtype=np.float64),
                   'nom_noise': np.zeros(len(t_value))}

    if gain_in['denom'][0] == 1:
        last_jd_var = ssm_ts['sm_jd'][0]
    else:
        last_jd_var = gain_in['last_jd']

    norm_factor = \
        np.sum(np.exp(-(np.tensordot(np.arange(1000, dtype=np.float32),
                                     (1. / t_value), axes=0))), axis=0)

    nom_noise = np.ones(len(t_value), dtype=np.float64)

    if 'sm_noise' in ssm_ts.dtype.fields:
        if gain_in['nom_noise'] is not None:
            nom_noise = gain_in['nom_noise']

        swi, swi_noise, nom, denom, last_jd_var, nom_ns = \
            swi_calc_cy_noise(ssm_ts['sm_jd'], ssm_ts['sm'], t_value, swi_jd,
                              gain_in['nom'], gain_in['denom'], last_jd_var,
                              ssm_ts['sm_noise'], nom_noise)
        swi_qflag = np.zeros_like(swi)
    else:
        swi, swi_qflag, nom, denom, last_jd_var = \
            swi_calc_cy(ssm_ts['sm_jd'], ssm_ts['sm'], t_value, swi_jd,
                        gain_in['nom'], gain_in['denom'], last_jd_var,
                        norm_factor, nan)
        swi_noise = np.zeros_like(swi)

    gain_out = {'denom': denom, 'nom': nom,
                'last_jd': last_jd_var, 'nom_noise': nom_noise}

    dtype_list = [('swi_jd', np.float64)]

    for t in t_value:
        dtype_list.append(('swi_{}'.format(t), np.float32))
        dtype_list.append(('swi_noise_{}'.format(t), np.float32))
        dtype_list.append(('swi_qflag_{}'.format(t), np.float32))

    swi_ts = np.zeros(swi_jd.size, dtype=np.dtype(dtype_list))

    swi_ts['swi_jd'] = swi_jd
    for i, t in enumerate(t_value):
        swi_ts['swi_{}'.format(t)] = swi[:, i]
        swi_ts['swi_noise_{}'.format(t)] = swi_noise[:, i]
        swi_ts['swi_qflag_{}'.format(t)] = swi_qflag[:, i]

    return swi_ts, gain_out

def calc_swi_noise_rec(ssm_ts, t_value, last_den=1, last_nom=0):
    """
    Recursive calculation of Soil Water Index (SWI) noise.

    Parameters
    ----------
    ssm_ts : numpy.ndarray
        Surface soil moisture time series with fields: sm_jd, sm, sm_noise
    t_value : numpy.ndarray
        Characteristic time length.
    denom : float
        denom value of the last calculation and starting point for
        the calculation.
    nom : float
        nom value of the last calculation and starting point for
        the calculation.

    Returns
    -------
    swi_noise_ts : numpy.ndarray
        Soil Water Index noise time series.
    """
    len_sm = len(ssm_ts)
    len_t_value = len(t_value)

    # var{n_swi} calculation
    nom_sn = np.zeros((len_sm, len_t_value))
    den = np.zeros((len_sm, len_t_value))
    swi_noise = np.zeros((len_sm, len_t_value))

    last_jd = ssm_ts['sm_jd'][0]
    den_values = np.ones(len_t_value)
    nom_values = np.zeros(len_t_value)

    den_values.fill(last_den)
    nom_values.fill(last_nom)

    for i in range(0, len_sm):
        for c in range(0, len_t_value):
            tdiff = (ssm_ts['sm_jd'][i]-last_jd) / t_value[c]
            fn = exp((-1) * tdiff)
            den[i][c] = 1 + fn * den_values[c]
            den_sn = den[i][c] ** 2

            exp_term = exp((-2)*tdiff)
            nom_sn[i][c] = nom_values[c] * exp_term + ssm_ts['sm_noise'][i]

            swi_noise[i][c] = nom_sn[i][c] / den_sn
            den_values[c] = den[i][c]
            nom_values[c] = nom_sn[i][c]
        last_jd = ssm_ts['sm_jd'][i]

    gain_out = {'denom': den_values, 'nom': nom_values, 'last_jd': last_jd,
                'nom_noise': 0}

    dtype_list = [('swi_jd', np.float64)]
    for t in t_value:
        dtype_list.append(('swi_noise_{}'.format(t), np.float32))

    swi_ts = np.zeros(ssm_ts.size, dtype=np.dtype(dtype_list))

    swi_ts['swi_jd'] = ssm_ts['sm_jd']
    for i, t in enumerate(t_value):
        swi_ts['swi_noise_{}'.format(t)] = swi_noise[:, i]

    return swi_ts, gain_out


# Copyright (c) 2019, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
This module tests the soil water index calculation from SSM time series.
"""

import numpy as np
from numpy.lib.recfunctions import unstructured_to_structured
import pandas as pd

from pyswi.swi_ts.swi_ts import calc_swi_ts, calc_swi_noise_rec


def test_process_swi_calc():
    """
    Test correct calculation of SWI, compared to a hardcoded calculation output.
    """
    t_value = [5, 50, 100]

    swi_jd = pd.date_range(
        '2007-01-01', periods=9).to_julian_date().values.astype(np.float64)

    sm = np.array([10, 20, 30, 40, 50, 60, 70, 80, 90], dtype=np.float32)

    dtype = np.dtype([('sm_jd', np.float64), ('sm', np.float32)])
    ssm_ts = unstructured_to_structured(
        np.hstack((swi_jd[:, np.newaxis], sm[:, np.newaxis])), dtype=dtype)

    swi_ts, gain_out = calc_swi_ts(ssm_ts, swi_jd, t_value=t_value)

    swi_ref = {
        'swi_5': np.array([10., 15.49834, 21.32452, 27.472094, 33.93228, 40.69421,
                           47.7452, 55.07107, 62.65647]),
        'swi_50': np.array([10., 15.049998, 20.133324, 25.249971, 30.399931, 35.58319,
                            40.799732, 46.049545, 51.332603]),
        'swi_100': np.array([10., 15.025, 20.066666, 25.124996, 30.199991, 35.29165,
                             40.399967, 45.524944, 50.666576])
    }

    for t in t_value:
        np.testing.assert_array_almost_equal(swi_ts['swi_{}'.format(t)],
                                             swi_ref['swi_{}'.format(t)], 4)

def test_process_swi_calc_nan_values():
    """
    Test correct calculation of SWI, compared to a hardcoded calculation output
    including nan values.
    """
    t_value = [5, 50, 100]

    swi_jd = pd.date_range(
        '2007-01-01', periods=9).to_julian_date().values.astype(np.float64)

    sm = np.array([10, 20, 30, 40, 50, 60, 999, 80, 10])

    dtype = np.dtype([('sm_jd', np.float64), ('sm', np.float32)])
    ssm_ts = unstructured_to_structured(
        np.hstack((swi_jd[:, np.newaxis], sm[:, np.newaxis])), dtype=dtype)

    swi_ts, gain_out = calc_swi_ts(ssm_ts, swi_jd, t_value=t_value, nan=999)

    swi_ref = {
        'swi_5': np.array([10., 15.49834, 21.32452, 27.472094, 33.93228, 40.69421,
                           np.nan, 51.660824, 41.072067]),
        'swi_50': np.array([10., 15.049998, 20.133324, 25.249971, 30.399931, 35.58319,
                            np.nan, 42.430466, 38.023155]),
        'swi_100': ([10., 15.025, 20.066666, 25.124996, 30.199991, 35.29165,
                     np.nan, 41.928066, 37.76523])
    }

    for t in t_value:
        np.testing.assert_array_almost_equal(swi_ts['swi_{}'.format(t)],
                                             swi_ref['swi_{}'.format(t)], 4)

def test_process_swi_gain():
    """
    Test gain in/out of the SWI calculation.
    It calculates the first 4 days and returns the value and gain and
    then starts the calculation of day 5 to 9 and
    compares it with hardcoded output over the whole period.
    """
    t_value = [5, 50, 100]

    swi_jd = pd.date_range(
        '2007-01-01', periods=9).to_julian_date().values.astype(np.float64)

    sm = np.array([10, 20, 30, 40, 50, 60, 999, 80, 10])

    dtype = np.dtype([('sm_jd', np.float64), ('sm', np.float32)])
    ssm_ts_part1 = unstructured_to_structured(
        np.hstack((swi_jd[:5, np.newaxis], sm[:5, np.newaxis])), dtype=dtype)
    ssm_ts_part2 = unstructured_to_structured(
        np.hstack((swi_jd[5:, np.newaxis], sm[5:, np.newaxis])), dtype=dtype)

    swi_ts_1, gain_1 = calc_swi_ts(ssm_ts_part1, swi_jd[:5],
                                   t_value=t_value, nan=999)

    swi_ts_2, gain_2 = calc_swi_ts(ssm_ts_part2, swi_jd[5:],
                                   t_value=t_value, gain_in=gain_1,
                                   nan=999)

    swi_ref = {
        'swi_5': np.array([10., 15.49834, 21.32452, 27.472094, 33.93228, 40.69421,
                           np.nan, 51.660824, 41.072067]),
        'swi_50': np.array([10., 15.049998, 20.133324, 25.249971, 30.399931, 35.58319,
                            np.nan, 42.430466, 38.023155]),
        'swi_100': ([10., 15.025, 20.066666, 25.124996, 30.199991, 35.29165,
                     np.nan, 41.928066, 37.76523])
    }

    for t in t_value:
        comb_swi_ts = swi_ts_1['swi_{}'.format(t)].tolist() + \
                      swi_ts_2['swi_{}'.format(t)].tolist()
        np.testing.assert_array_almost_equal(comb_swi_ts, swi_ref['swi_{}'.format(t)], 4)

def test_process_swi_not_daily_out():
    """
    Test correct calculation of SWI with irregular timestamps,
    compared to the a hardcoded calculation output.
    """
    t_value = [5, 50, 100]

    swi_jd = np.array([2454102.3093969999, 2454103.3644969999,
                       2454103.873199, 2454104.9282769999,
                       2454105.9139760002, 2454106.3213539999,
                       2454107.3070530002, 2454108.3621530002,
                       2454108.8708330002, 2454109.9259549999])

    sm = np.array([84, 88, 79, 87, 91, 93, 92, 85, 85, 90])

    dtype = np.dtype([('sm_jd', np.float64), ('sm', np.float32)])
    ssm_ts = unstructured_to_structured(
        np.hstack((swi_jd[:, np.newaxis], sm[:, np.newaxis])), dtype=dtype)

    swi_ts, gain_out = calc_swi_ts(ssm_ts, swi_jd, t_value=t_value)

    swi_ref = {
        'swi_5': np.array([84., 86.21024, 83.47359, 84.59898, 86.39057, 87.93006,
                           88.829475, 88.008446, 87.43135, 87.9233]),
        'swi_50': np.array([84., 86.0211, 83.64838, 84.50836, 85.854836, 87.090576,
                            87.826706, 87.45131, 87.16135, 87.46732]),
        'swi_100': np.array([84., 86.01055, 83.65755, 84.50414, 85.827286, 87.04517,
                             87.77045, 87.41343, 87.13668, 87.434044])
    }

    for t in t_value:
        np.testing.assert_array_almost_equal(swi_ts['swi_{}'.format(t)],
                                             swi_ref['swi_{}'.format(t)], 4)

def test_process_swi_noise():
    """
    Test correct calculation of SWI Noise inside the SWI calculation
    process.
    """
    t_value = [5, 50, 100]

    swi_jd = pd.date_range(
        '2007-01-01', periods=9).to_julian_date().values.astype(np.float64)

    sm_noise = np.zeros(9)
    sm_noise.fill(3)

    sm = np.array([10, 20, 30, 40, 50, 60, 70, 80, 90])

    dtype = np.dtype([('sm_jd', np.float64), ('sm', np.float32),
                      ('sm_noise', np.float32)])
    ssm_ts = unstructured_to_structured(
        np.hstack((swi_jd[:, np.newaxis], sm[:, np.newaxis],
                   sm_noise[:, np.newaxis])), dtype=dtype)

    swi_ts1, _ = calc_swi_ts(ssm_ts, swi_jd, t_value=t_value,
                             nom_init=1, denom_init=1)

    swi_ts2, _ = calc_swi_noise_rec(ssm_ts, t_value)

    for t in t_value:
        np.testing.assert_array_almost_equal(
            swi_ts1['swi_noise_{}'.format(t)],
            swi_ts2['swi_noise_{}'.format(t)], 4)

def test_swi_gain_noise():
    """
    Test correct calculation of SWI noise gain.
    """
    t_value = [5, 50, 100]

    swi_jd = pd.date_range('2007-01-01', periods=9).to_julian_date().values
    sm = np.array([10, 20, 30, 40, 50, 60, 70, 80, 90])
    sm_noise = np.zeros(9)
    sm_noise.fill(3)

    dtype = np.dtype([('sm_jd', np.float64), ('sm', np.float32),
                      ('sm_noise', np.float32)])
    ssm_ts = unstructured_to_structured(
        np.hstack((swi_jd[:, np.newaxis], sm[:, np.newaxis],
                   sm_noise[:, np.newaxis])), dtype=dtype)

    swi_ts_all, gain_all = calc_swi_ts(ssm_ts, swi_jd, t_value=t_value)
    swi_ts_1, gain_1 = calc_swi_ts(ssm_ts[:5], swi_jd[:5], t_value=t_value)
    swi_ts_2, gain_2 = calc_swi_ts(ssm_ts[5:], swi_jd[5:],
                                   t_value=t_value, gain_in=gain_1)

    for t in t_value:
        comb_swi_ts = swi_ts_1['swi_noise_{}'.format(t)].tolist() + \
            swi_ts_2['swi_noise_{}'.format(t)].tolist()
        np.testing.assert_array_almost_equal(
            comb_swi_ts, swi_ts_all['swi_noise_{}'.format(t)], 4)
